@@ -13,13 +13,22 @@ We can also move the implementation to C++ if desired.
 """
 
 import inspect
+import shutil
 from collections.abc import KeysView
 from io import BytesIO
+from pathlib import Path
 from subprocess import PIPE, run
 from tempfile import NamedTemporaryFile
 
 from . import Array, Dictionary, Name, Object, Page, Pdf, Stream
-from ._qpdf import PdfError, StreamParser, Token, _ObjectMapping
+from ._qpdf import (
+    AccessMode,
+    ObjectStreamMode,
+    PdfError,
+    StreamParser,
+    Token,
+    _ObjectMapping,
+)
 from .models import EncryptionInfo, Outline, PdfMetadata, Permissions
 
 # pylint: disable=no-member,unsupported-membership-test,unsubscriptable-object
@@ -55,10 +64,13 @@ def augments(cls_cpp):
     Any existing methods may be used, regardless of whether they defined
     elsewhere in the support class or in the target class.
 
-    The target class does not have to be C++ or derived from pybind11.
+    For data fields to work, including @property accessors, the target class must be
+    tagged ``py::dynamic_attr`` in pybind11.
 
-    THIS DOES NOT work for static methods or class methods. pybind11 does not seem
-    to support this sort of runtime modification.
+    Strictly, the target class does not have to be C++ or derived from pybind11.
+    This works on pure Python classes too.
+
+    THIS DOES NOT work for class methods.
     """
 
     def class_augment(cls, cls_cpp=cls_cpp):
@@ -259,10 +271,17 @@ class Extend_Pdf:
         self, set_pikepdf_as_editor=True, update_docinfo=True, strict=False
     ):
         """
-        Open the PDF's XMP metadata for editing
+        Open the PDF's XMP metadata for editing.
 
-        Recommend for use in a ``with`` block. Changes are committed to the
-        PDF when the block exits. (The ``Pdf`` must still be opened.)
+        There is no ``.close()`` function on the metadata object, since this is
+        intended to be used inside a ``with`` block only.
+
+        For historical reasons, certain parts of PDF metadata are stored in
+        two different locations and formats. This feature coordinates edits so
+        that both types of metadata are updated consistently and "atomically"
+        (assuming single threaded access). It operates on the ``Pdf`` in memory,
+        not any file on disk. To persist metadata changes, you must still use
+        ``Pdf.save()``.
 
         Example:
             >>> with pdf.open_metadata() as meta:
@@ -418,6 +437,8 @@ class Extend_Pdf:
         else:
             description = "closed object"
         self._process(description, EMPTY_PDF)
+        if getattr(self, '_tmp_stream', None):
+            self._tmp_stream.close()
 
     def __enter__(self):
         return self
@@ -564,6 +585,261 @@ class Extend_Pdf:
         if '/PageMode' not in self.Root:
             self.Root.PageMode = Name.UseAttachments
 
+    def save(
+        self,
+        filename_or_stream=None,
+        static_id=False,
+        preserve_pdfa=True,
+        min_version="",
+        force_version="",
+        fix_metadata_version=True,
+        compress_streams=True,
+        stream_decode_level=None,
+        object_stream_mode=ObjectStreamMode.preserve,
+        normalize_content=False,
+        linearize=False,
+        qdf=False,
+        progress=None,
+        encryption=None,
+    ):
+        """
+        Save all modifications to this :class:`pikepdf.Pdf`.
+
+        Args:
+            filename (Path or str or stream): Where to write the output. If a file
+                exists in this location it will be overwritten.
+                If the file was opened with ``allow_overwriting_input=True``,
+                then it is permitted to overwrite the original file, and
+                this parameter may be omitted to implicitly use the original
+                filename. Otherwise, the filename may not be the same as the
+                input file, as overwriting the input file would corrupt data
+                since pikepdf using lazy loading.
+
+            static_id (bool): Indicates that the ``/ID`` metadata, normally
+                calculated as a hash of certain PDF contents and metadata
+                including the current time, should instead be generated
+                deterministically. Normally for debugging.
+            preserve_pdfa (bool): Ensures that the file is generated in a
+                manner compliant with PDF/A and other stricter variants.
+                This should be True, the default, in most cases.
+
+            min_version (str or tuple): Sets the minimum version of PDF
+                specification that should be required. If left alone QPDF
+                will decide. If a tuple, the second element is an integer, the
+                extension level. If the version number is not a valid format,
+                QPDF will decide what to do.
+            force_version (str or tuple): Override the version recommend by QPDF,
+                potentially creating an invalid file that does not display
+                in old versions. See QPDF manual for details. If a tuple, the
+                second element is an integer, the extension level.
+            fix_metadata_version (bool): If ``True`` (default) and the XMP metadata
+                contains the optional PDF version field, ensure the version in
+                metadata is correct. If the XMP metadata does not contain a PDF
+                version field, none will be added. To ensure that the field is
+                added, edit the metadata and insert a placeholder value in
+                ``pdf:PDFVersion``. If XMP metadata does not exist, it will
+                not be created regardless of the value of this argument.
+
+            object_stream_mode (pikepdf.ObjectStreamMode):
+                ``disable`` prevents the use of object streams.
+                ``preserve`` keeps object streams from the input file.
+                ``generate`` uses object streams wherever possible,
+                creating the smallest files but requiring PDF 1.5+.
+
+            compress_streams (bool): Enables or disables the compression of
+                stream objects in the PDF. Metadata is never compressed.
+                By default this is set to ``True``, and should be except
+                for debugging.
+
+            stream_decode_level (pikepdf.StreamDecodeLevel): Specifies how
+                to encode stream objects. See documentation for
+                ``StreamDecodeLevel``.
+
+            normalize_content (bool): Enables parsing and reformatting the
+                content stream within PDFs. This may debugging PDFs easier.
+
+            linearize (bool): Enables creating linear or "fast web view",
+                where the file's contents are organized sequentially so that
+                a viewer can begin rendering before it has the whole file.
+                As a drawback, it tends to make files larger.
+
+            qdf (bool): Save output QDF mode.  QDF mode is a special output
+                mode in QPDF to allow editing of PDFs in a text editor. Use
+                the program ``fix-qdf`` to fix convert back to a standard
+                PDF.
+
+            progress (callable): Specify a callback function that is called
+                as the PDF is written. The function will be called with an
+                integer between 0-100 as the sole parameter, the progress
+                percentage. This function may not access or modify the PDF
+                while it is being written, or data corruption will almost
+                certainly occur.
+
+            encryption (pikepdf.models.Encryption or bool): If ``False``
+                or omitted, existing encryption will be removed. If ``True``
+                encryption settings are copied from the originating PDF.
+                Alternately, an ``Encryption`` object may be provided that
+                sets the parameters for new encryption.
+
+        You may call ``.save()`` multiple times with different parameters
+        to generate different versions of a file, and you *may* continue
+        to modify the file after saving it. ``.save()`` does not modify
+        the ``Pdf`` object in memory, except possibly by updating the XMP
+        metadata version with ``fix_metadata_version``.
+
+        .. note::
+
+            :meth:`pikepdf.Pdf.remove_unreferenced_resources` before saving
+            may eliminate unnecessary resources from the output file, so
+            calling this method before saving is recommended. This is not
+            done automatically because ``.save()`` is intended to be
+            idempotent.
+
+        .. note::
+
+            pikepdf can read PDFs will incremental updates, but always
+            any coalesces incremental updates into a single non-incremental
+            PDF file when saving.
+        """
+        if not filename_or_stream and self._original_filename:
+            filename_or_stream = self._original_filename
+        self._save(
+            filename_or_stream,
+            static_id=static_id,
+            preserve_pdfa=preserve_pdfa,
+            min_version=min_version,
+            force_version=force_version,
+            fix_metadata_version=fix_metadata_version,
+            compress_streams=compress_streams,
+            stream_decode_level=stream_decode_level,
+            object_stream_mode=object_stream_mode,
+            normalize_content=normalize_content,
+            linearize=linearize,
+            qdf=qdf,
+            progress=progress,
+            encryption=encryption,
+            samefile_check=getattr(self, '_tmp_stream', None) is None,
+        )
+
+    @staticmethod
+    def open(
+        filename_or_stream,
+        password="",
+        hex_password=False,
+        ignore_xref_streams=False,
+        suppress_warnings=True,
+        attempt_recovery=True,
+        inherit_page_attributes=True,
+        access_mode=AccessMode.default,
+        allow_overwriting_input=False,
+    ):
+        """
+        Open an existing file at *filename_or_stream*.
+
+        If *filename_or_stream* is path-like, the file will be opened for reading.
+        The file should not be modified by another process while it is open in
+        pikepdf, or undefined behavior may occur. This is because the file may be
+        lazily loaded. Despite this restriction, pikepdf does not try to use any OS
+        services to obtain an exclusive lock on the file. Some applications may
+        want to attempt this or copy the file to a temporary location before
+        editing. This behaviour change if *allow_overwriting_input* is set: the whole
+        file is then read and copied to memory, so that pikepdf can overwrite it
+        when calling ``.save()``.
+
+        When this is function is called with a stream-like object, you must ensure
+        that the data it returns cannot be modified, or undefined behavior will
+        occur.
+
+        Any changes to the file must be persisted by using ``.save()``.
+
+        If *filename_or_stream* has ``.read()`` and ``.seek()`` methods, the file
+        will be accessed as a readable binary stream. pikepdf will read the
+        entire stream into a private buffer.
+
+        ``.open()`` may be used in a ``with``-block; ``.close()`` will be called when
+        the block exits, if applicable.
+
+        Whenever pikepdf opens a file, it will close it. If you open the file
+        for pikepdf or give it a stream-like object to read from, you must
+        release that object when appropriate.
+
+        Examples:
+
+            >>> with Pdf.open("test.pdf") as pdf:
+                    ...
+
+            >>> pdf = Pdf.open("test.pdf", password="rosebud")
+
+        Args:
+            filename_or_stream (os.PathLike): Filename of PDF to open.
+            password (str or bytes): User or owner password to open an
+                encrypted PDF. If the type of this parameter is ``str``
+                it will be encoded as UTF-8. If the type is ``bytes`` it will
+                be saved verbatim. Passwords are always padded or
+                truncated to 32 bytes internally. Use ASCII passwords for
+                maximum compatibility.
+            hex_password (bool): If True, interpret the password as a
+                hex-encoded version of the exact encryption key to use, without
+                performing the normal key computation. Useful in forensics.
+            ignore_xref_streams (bool): If True, ignore cross-reference
+                streams. See qpdf documentation.
+            suppress_warnings (bool): If True (default), warnings are not
+                printed to stderr. Use :meth:`pikepdf.Pdf.get_warnings()` to
+                retrieve warnings.
+            attempt_recovery (bool): If True (default), attempt to recover
+                from PDF parsing errors.
+            inherit_page_attributes (bool): If True (default), push attributes
+                set on a group of pages to individual pages
+            access_mode (pikepdf.AccessMode): If ``.default``, pikepdf will
+                decide how to access the file. Currently, it will always
+                selected stream access. To attempt memory mapping and fallback
+                to stream if memory mapping failed, use ``.mmap``.  Use
+                ``.mmap_only`` to require memory mapping or fail
+                (this is expected to only be useful for testing). Applications
+                should be prepared to handle the SIGBUS signal on POSIX in
+                the event that the file is successfully mapped but later goes
+                away.
+            allow_overwriting_input (bool): If True, allows calling ``.save()``
+                to overwrite the input file. This is performed by loading the
+                entire input file into memory at open time; this will use more
+                memory and may recent performance especially when the opened
+                file will not be modified.
+        Raises:
+            pikepdf.PasswordError: If the password failed to open the
+                file.
+            pikepdf.PdfError: If for other reasons we could not open
+                the file.
+            TypeError: If the type of ``filename_or_stream`` is not
+                usable.
+            FileNotFoundError: If the file was not found.
+        """
+        tmp_stream, original_filename = None, False
+        if allow_overwriting_input:
+            try:
+                Path(filename_or_stream)
+            except TypeError as error:
+                raise ValueError(
+                    '"allow_overwriting_input=True" requires "open" first argument '
+                    'to be a file path'
+                ) from error
+            original_filename = str(filename_or_stream)
+            with open(original_filename, 'rb') as pdf_file:
+                tmp_stream = BytesIO()
+                shutil.copyfileobj(pdf_file, tmp_stream)
+        pdf = Pdf._open(
+            tmp_stream or filename_or_stream,
+            password=password,
+            hex_password=hex_password,
+            ignore_xref_streams=ignore_xref_streams,
+            suppress_warnings=suppress_warnings,
+            attempt_recovery=attempt_recovery,
+            inherit_page_attributes=inherit_page_attributes,
+            access_mode=access_mode,
+        )
+        setattr(pdf, '_tmp_stream', tmp_stream)
+        setattr(pdf, '_original_filename', original_filename)
+        return pdf
+
 
 @augments(_ObjectMapping)
 class Extend_ObjectMapping:
